"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisTrigger = void 0;
const n8n_workflow_1 = require("n8n-workflow");
const redis_1 = __importDefault(require("redis"));
class RedisTrigger {
    constructor() {
        this.description = {
            displayName: 'Redis Trigger',
            name: 'redisTrigger',
            icon: 'file:redis.svg',
            group: ['trigger'],
            version: 1,
            description: 'Subscribe to redis channel',
            defaults: {
                name: 'Redis Trigger',
            },
            inputs: [],
            outputs: ['main'],
            credentials: [
                {
                    name: 'redis',
                    required: true,
                },
            ],
            properties: [
                {
                    displayName: 'Channels',
                    name: 'channels',
                    type: 'string',
                    default: '',
                    required: true,
                    description: 'Channels to subscribe to, multiple channels be defined with comma. Wildcard character(*) is supported.',
                },
                {
                    displayName: 'Options',
                    name: 'options',
                    type: 'collection',
                    placeholder: 'Add Option',
                    default: {},
                    options: [
                        {
                            displayName: 'JSON Parse Body',
                            name: 'jsonParseBody',
                            type: 'boolean',
                            default: false,
                            description: 'Whether to try to parse the message to an object',
                        },
                        {
                            displayName: 'Only Message',
                            name: 'onlyMessage',
                            type: 'boolean',
                            default: false,
                            description: 'Whether to return only the message property',
                        },
                    ],
                },
            ],
        };
    }
    async trigger() {
        const credentials = await this.getCredentials('redis');
        const redisOptions = {
            host: credentials.host,
            port: credentials.port,
            db: credentials.database,
        };
        if (credentials.password) {
            redisOptions.password = credentials.password;
        }
        const channels = this.getNodeParameter('channels').split(',');
        const options = this.getNodeParameter('options');
        if (!channels) {
            throw new n8n_workflow_1.NodeOperationError(this.getNode(), 'Channels are mandatory!');
        }
        const client = redis_1.default.createClient(redisOptions);
        const manualTriggerFunction = async () => {
            await new Promise((resolve, reject) => {
                client.on('connect', () => {
                    for (const channel of channels) {
                        client.psubscribe(channel);
                    }
                    client.on('pmessage', (pattern, channel, message) => {
                        if (options.jsonParseBody) {
                            try {
                                message = JSON.parse(message);
                            }
                            catch (error) { }
                        }
                        if (options.onlyMessage) {
                            this.emit([this.helpers.returnJsonArray({ message })]);
                            resolve(true);
                            return;
                        }
                        this.emit([this.helpers.returnJsonArray({ channel, message })]);
                        resolve(true);
                    });
                });
                client.on('error', (error) => {
                    reject(error);
                });
            });
        };
        if (this.getMode() === 'trigger') {
            await manualTriggerFunction();
        }
        async function closeFunction() {
            client.quit();
        }
        return {
            closeFunction,
            manualTriggerFunction,
        };
    }
}
exports.RedisTrigger = RedisTrigger;
//# sourceMappingURL=RedisTrigger.node.js.map