"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getWorkflowFilenames = exports.testWorkflows = exports.workflowToTests = exports.equalityTest = exports.readJsonFileSync = exports.getResultNodeData = exports.setup = exports.initBinaryDataManager = exports.createTemporaryDir = exports.NodeTypes = exports.WorkflowExecuteAdditionalData = exports.CredentialsHelper = void 0;
const fs_1 = require("fs");
const n8n_core_1 = require("n8n-core");
const n8n_workflow_1 = require("n8n-workflow");
const ExecuteWorkflow_1 = require("./ExecuteWorkflow");
const path_1 = __importDefault(require("path"));
const os_1 = require("os");
const lodash_1 = require("lodash");
class CredentialsHelper extends n8n_workflow_1.ICredentialsHelper {
    async authenticate(credentials, typeName, requestParams) {
        return requestParams;
    }
    async preAuthentication(helpers, credentials, typeName, node, credentialsExpired) {
        return undefined;
    }
    getParentTypes(name) {
        return [];
    }
    async getDecrypted(nodeCredentials, type) {
        return {};
    }
    async getCredentials(nodeCredentials, type) {
        return new n8n_core_1.Credentials({ id: null, name: '' }, '', [], '');
    }
    async updateCredentials(nodeCredentials, type, data) { }
}
exports.CredentialsHelper = CredentialsHelper;
function WorkflowExecuteAdditionalData(waitPromise, nodeExecutionOrder) {
    const hookFunctions = {
        nodeExecuteAfter: [
            async (nodeName, data) => {
                nodeExecutionOrder.push(nodeName);
            },
        ],
        workflowExecuteAfter: [
            async (fullRunData) => {
                waitPromise.resolve(fullRunData);
            },
        ],
    };
    const workflowData = {
        name: '',
        createdAt: new Date(),
        updatedAt: new Date(),
        active: true,
        nodes: [],
        connections: {},
    };
    return {
        credentialsHelper: new CredentialsHelper(''),
        hooks: new n8n_workflow_1.WorkflowHooks(hookFunctions, 'trigger', '1', workflowData),
        executeWorkflow: async (workflowInfo) => { },
        sendMessageToUI: (message) => { },
        restApiUrl: '',
        encryptionKey: 'test',
        timezone: 'America/New_York',
        webhookBaseUrl: 'webhook',
        webhookWaitingBaseUrl: 'webhook-waiting',
        webhookTestBaseUrl: 'webhook-test',
        userId: '123',
    };
}
exports.WorkflowExecuteAdditionalData = WorkflowExecuteAdditionalData;
class NodeTypesClass {
    constructor() {
        this.nodeTypes = {};
    }
    getByName(nodeType) {
        return this.nodeTypes[nodeType].type;
    }
    addNode(nodeTypeName, nodeType) {
        const loadedNode = {
            [nodeTypeName]: {
                sourcePath: '',
                type: nodeType,
            },
        };
        this.nodeTypes = {
            ...this.nodeTypes,
            ...loadedNode,
        };
    }
    getByNameAndVersion(nodeType, version) {
        return n8n_workflow_1.NodeHelpers.getVersionedNodeType(this.nodeTypes[nodeType].type, version);
    }
}
let nodeTypesInstance;
function NodeTypes() {
    if (nodeTypesInstance === undefined) {
        nodeTypesInstance = new NodeTypesClass();
    }
    return nodeTypesInstance;
}
exports.NodeTypes = NodeTypes;
let knownNodes = null;
const loadKnownNodes = () => {
    if (knownNodes === null) {
        knownNodes = JSON.parse((0, fs_1.readFileSync)('dist/known/nodes.json').toString());
    }
    return knownNodes;
};
function createTemporaryDir(prefix = 'n8n') {
    return (0, fs_1.mkdtempSync)(path_1.default.join((0, os_1.tmpdir)(), prefix));
}
exports.createTemporaryDir = createTemporaryDir;
async function initBinaryDataManager(mode = 'default') {
    const temporaryDir = createTemporaryDir();
    await n8n_core_1.BinaryDataManager.init({
        mode,
        availableModes: mode,
        localStoragePath: temporaryDir,
        binaryDataTTL: 1,
        persistedBinaryDataTTL: 1,
    });
    return temporaryDir;
}
exports.initBinaryDataManager = initBinaryDataManager;
function setup(testData) {
    if (!Array.isArray(testData)) {
        testData = [testData];
    }
    const knownNodes = loadKnownNodes();
    const nodeTypes = NodeTypes();
    const nodeNames = Array.from(new Set(testData.flatMap((data) => data.input.workflowData.nodes.map((n) => n.type))));
    for (const nodeName of nodeNames) {
        if (!nodeName.startsWith('n8n-nodes-base.')) {
            throw new Error(`Unknown node type: ${nodeName}`);
        }
        const loadInfo = knownNodes[nodeName.replace('n8n-nodes-base.', '')];
        if (!loadInfo) {
            throw new Error(`Unknown node type: ${nodeName}`);
        }
        const node = (0, n8n_core_1.loadClassInIsolation)(path_1.default.join(process.cwd(), loadInfo.sourcePath), loadInfo.className);
        nodeTypes.addNode(nodeName, node);
    }
    const fakeLogger = {
        log: () => { },
        debug: () => { },
        verbose: () => { },
        info: () => { },
        warn: () => { },
        error: () => { },
    };
    n8n_workflow_1.LoggerProxy.init(fakeLogger);
    return nodeTypes;
}
exports.setup = setup;
function getResultNodeData(result, testData) {
    return Object.keys(testData.output.nodeData).map((nodeName) => {
        if (result.data.resultData.runData[nodeName] === undefined) {
            throw new Error(`Data for node "${nodeName}" is missing!`);
        }
        const resultData = result.data.resultData.runData[nodeName].map((nodeData) => {
            if (nodeData.data === undefined) {
                return null;
            }
            return nodeData.data.main[0].map((entry) => {
                if (entry.binary && (0, lodash_1.isEmpty)(entry.binary))
                    delete entry.binary;
                delete entry.pairedItem;
                return entry;
            });
        });
        return {
            nodeName,
            resultData,
        };
    });
}
exports.getResultNodeData = getResultNodeData;
function readJsonFileSync(path) {
    return JSON.parse((0, fs_1.readFileSync)(path, 'utf-8'));
}
exports.readJsonFileSync = readJsonFileSync;
const equalityTest = async (testData, types) => {
    const { result } = await (0, ExecuteWorkflow_1.executeWorkflow)(testData, types);
    const resultNodeData = getResultNodeData(result, testData);
    resultNodeData.forEach(({ nodeName, resultData }) => {
        return expect(resultData).toEqual(testData.output.nodeData[nodeName]);
    });
    expect(result.finished).toEqual(true);
};
exports.equalityTest = equalityTest;
const preparePinData = (pinData) => {
    const returnData = Object.keys(pinData).reduce((acc, key) => {
        const data = pinData[key];
        acc[key] = [data];
        return acc;
    }, {});
    return returnData;
};
const workflowToTests = (workflowFiles) => {
    const testCases = [];
    for (const filePath of workflowFiles) {
        const description = filePath.replace('.json', '');
        const workflowData = readJsonFileSync(filePath);
        if (workflowData.pinData === undefined) {
            throw new Error('Workflow data does not contain pinData');
        }
        const nodeData = preparePinData(workflowData.pinData);
        delete workflowData.pinData;
        const input = { workflowData };
        const output = { nodeData };
        testCases.push({ description, input, output });
    }
    return testCases;
};
exports.workflowToTests = workflowToTests;
const testWorkflows = (workflows) => {
    const tests = (0, exports.workflowToTests)(workflows);
    const nodeTypes = setup(tests);
    for (const testData of tests) {
        test(testData.description, async () => (0, exports.equalityTest)(testData, nodeTypes));
    }
};
exports.testWorkflows = testWorkflows;
const getWorkflowFilenames = (dirname) => {
    const workflows = [];
    const filenames = (0, fs_1.readdirSync)(dirname);
    const testFolder = dirname.split(`${path_1.default.sep}nodes-base${path_1.default.sep}`)[1];
    filenames.forEach((file) => {
        if (file.includes('.json')) {
            workflows.push(path_1.default.join(testFolder, file));
        }
    });
    return workflows;
};
exports.getWorkflowFilenames = getWorkflowFilenames;
//# sourceMappingURL=Helpers.js.map